import json
import logging
import os
import sys
from typing import Dict, List

import click
import rich.progress

from biolib._data_record.data_record import DataRecord
from biolib.biolib_api_client import BiolibApiClient
from biolib.biolib_logging import logger, logger_no_user_data
from biolib.typing_utils import Optional


@click.group(help='Data Records')
def data_record() -> None:
    logger.configure(default_log_level=logging.INFO)
    logger_no_user_data.configure(default_log_level=logging.INFO)


@data_record.command(help='Create a Data Record')
@click.argument('uri', required=True)
@click.option('--data-path', required=True, type=click.Path(exists=True))
@click.option('--record-type', required=False, type=str, default=None)
def create(uri: str, data_path: str, record_type: Optional[str]) -> None:
    DataRecord.create(destination=uri, data_path=data_path, record_type=record_type)


@data_record.command(help='Update a Data Record')
@click.argument('uri', required=True)
@click.option('--data-path', required=True, type=click.Path(exists=True))
@click.option('--chunk-size', default=None, required=False, type=click.INT, help='The size of each chunk (In MB)')
def update(uri: str, data_path: str, chunk_size: Optional[int]) -> None:
    DataRecord.get_by_uri(uri=uri).update(data_path=data_path, chunk_size_in_mb=chunk_size)


@data_record.command(help='Download files from a Data Record')
@click.argument('uri', required=True)
@click.option('--file', required=False, type=str)
@click.option('--path-filter', required=False, type=str, hide_input=True)
def download(uri: str, file: Optional[str], path_filter: Optional[str]) -> None:
    record = DataRecord.get_by_uri(uri=uri)
    if file is not None:
        try:
            file_obj = [file_obj for file_obj in record.list_files() if file_obj.path == file][0]
        except IndexError:
            raise Exception('File not found in data record') from None

        assert not os.path.exists(file_obj.name), 'File already exists in current directory'
        with open(file_obj.name, 'wb') as file_handle:
            file_handle.write(file_obj.get_data())

    else:
        assert not os.path.exists(record.name), f'Directory with name {record.name} already exists in current directory'
        record.save_files(output_dir=record.name, path_filter=path_filter)


@data_record.command(help='Describe a Data Record')
@click.argument('uri', required=True)
@click.option('--json', 'output_as_json', is_flag=True, default=False, required=False, help='Format output as JSON')
def describe(uri: str, output_as_json: bool) -> None:
    BiolibApiClient.assert_is_signed_in(authenticated_action_description='get Data Record description')
    record = DataRecord.get_by_uri(uri)
    files_info: List[Dict] = []
    total_size_in_bytes = 0
    for file in record.list_files():
        files_info.append({'path': file.path, 'size_bytes': file.length})
        total_size_in_bytes += file.length

    if output_as_json:
        print(
            json.dumps(
                obj={'uri': record.uri, 'size_bytes': total_size_in_bytes, 'files': files_info},
                indent=4,
            )
        )
    else:
        print(f'Data Record {record.uri}\ntotal {total_size_in_bytes} bytes\n')
        print('size bytes    path')
        for file_info in files_info:
            size_string = str(file_info['size_bytes'])
            leading_space_string = ' ' * (10 - len(size_string))
            print(f"{leading_space_string}{size_string}    {file_info['path']}")


@data_record.command(help='Delete a Data Record')
@click.argument('uri', required=True)
def delete(uri: str) -> None:
    record = DataRecord.get_by_uri(uri=uri)

    print(f'You are about to delete the data record: {record.uri}')
    print('This action cannot be undone.')

    confirmation = input(f'To confirm deletion, please type the data record name "{record.name}": ')
    if confirmation != record.name:
        print('Data record name does not match. Deletion cancelled.')
        return

    record.delete()
    print(f'Data record {record.uri} has been deleted.')


def _clone_data_record_with_progress(
    source_record: DataRecord,
    dest_record: DataRecord,
) -> None:
    # pylint: disable=protected-access
    total_size_in_bytes = source_record._get_zip_size_bytes()
    # pylint: enable=protected-access

    if total_size_in_bytes == 0:
        logger.info('Source data record has no data to clone')
        return

    if sys.stdout.isatty():
        with rich.progress.Progress(
            rich.progress.TextColumn('[bold blue]{task.description}'),
            rich.progress.BarColumn(),
            rich.progress.TaskProgressColumn(),
            rich.progress.TimeRemainingColumn(),
            rich.progress.TransferSpeedColumn(),
        ) as progress:
            task_id = progress.add_task('Cloning data record', total=total_size_in_bytes)

            def on_progress(bytes_uploaded: int, _total_bytes: int) -> None:
                progress.update(task_id, completed=bytes_uploaded)

            DataRecord.clone(source=source_record, destination=dest_record, on_progress=on_progress)
    else:
        logger.info(f'Cloning ~{round(total_size_in_bytes / 10**6)}mb of data')
        DataRecord.clone(source=source_record, destination=dest_record)


def _get_or_create_destination_record(destination_uri: str) -> Optional[DataRecord]:
    try:
        return DataRecord.get_by_uri(uri=destination_uri)
    except Exception:
        print(f'Destination data record "{destination_uri}" does not exist.')
        confirmation = input('Would you like to create it? [y/N]: ')
        if confirmation.lower() != 'y':
            print('Clone cancelled.')
            return None

        return DataRecord.create(destination=destination_uri)


@data_record.command(help='Clone a Data Record to another location')
@click.argument('source_uri', required=True)
@click.argument('destination_uri', required=True)
def clone(source_uri: str, destination_uri: str) -> None:
    BiolibApiClient.assert_is_signed_in(authenticated_action_description='clone a Data Record')

    logger.info(f'Fetching source data record: {source_uri}')
    source_record = DataRecord.get_by_uri(uri=source_uri)

    logger.info(f'Checking destination data record: {destination_uri}')
    dest_record = _get_or_create_destination_record(destination_uri)
    if dest_record is None:
        return

    logger.info(f'Cloning from {source_record.uri} to {dest_record.uri}...')
    _clone_data_record_with_progress(source_record=source_record, dest_record=dest_record)
    logger.info('Clone completed successfully.')



import contextlib
import json
import os
import socket
import time
from typing import List, Optional

from docker.errors import NotFound
from docker.models.networks import Network

from biolib import utils
from biolib.biolib_docker_client import BiolibDockerClient
from biolib.biolib_logging import logger_no_user_data
from biolib.compute_node.job_worker.network_alloc import _allocate_network_with_retries


class NetworkBuffer:
    BUFFER_SIZE = 25
    NETWORK_NAME_PREFIX = 'biolib-remote-host-network-'
    NETWORK_LABEL = 'biolib-role=remote-host-network'

    _BIOLIB_DIR = '/biolib' if utils.IS_RUNNING_IN_CLOUD else '/tmp/biolib'
    _NETWORKS_FILE = os.path.join(_BIOLIB_DIR, 'remote-host-networks.json')
    _LOCK_FILE = os.path.join(_BIOLIB_DIR, 'remote-host-networks.lock')
    _LOCK_TIMEOUT_SECONDS = 60
    _STALE_LOCK_THRESHOLD_SECONDS = 600

    _instance: Optional['NetworkBuffer'] = None

    def __init__(self):
        os.makedirs(self._BIOLIB_DIR, exist_ok=True)
        self._docker = BiolibDockerClient.get_docker_client()

    @classmethod
    def get_instance(cls) -> 'NetworkBuffer':
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def _acquire_lock(self) -> None:
        start_time = time.time()
        retry_count = 0

        while time.time() - start_time < self._LOCK_TIMEOUT_SECONDS:
            try:
                with open(self._LOCK_FILE, 'x') as lock_file:
                    lock_info = {
                        'pid': os.getpid(),
                        'hostname': socket.gethostname(),
                        'started_at': time.time(),
                    }
                    json.dump(lock_info, lock_file)
                return
            except FileExistsError:
                if retry_count == 0:
                    self._check_and_remove_stale_lock()

                time.sleep(0.5)
                retry_count += 1

        raise RuntimeError(
            f'Failed to acquire network buffer lock after {self._LOCK_TIMEOUT_SECONDS}s: {self._LOCK_FILE}'
        )

    def _check_and_remove_stale_lock(self) -> None:
        try:
            if not os.path.exists(self._LOCK_FILE):
                return

            lock_mtime = os.path.getmtime(self._LOCK_FILE)
            lock_age = time.time() - lock_mtime

            if lock_age > self._STALE_LOCK_THRESHOLD_SECONDS:
                try:
                    with open(self._LOCK_FILE) as f:
                        lock_info = json.load(f)
                        lock_pid = lock_info.get('pid')

                        if lock_pid:
                            try:
                                os.kill(lock_pid, 0)
                                logger_no_user_data.warning(
                                    f'Lock file is old ({lock_age:.0f}s) but process {lock_pid} is still alive'
                                )
                                return
                            except (OSError, ProcessLookupError):
                                pass

                except (json.JSONDecodeError, KeyError, ValueError):
                    pass

                logger_no_user_data.warning(
                    f'Removing stale lock file (age: {lock_age:.0f}s, threshold: {self._STALE_LOCK_THRESHOLD_SECONDS}s)'
                )
                os.remove(self._LOCK_FILE)

        except Exception as error:
            logger_no_user_data.debug(f'Error checking stale lock: {error}')

    def _release_lock(self) -> None:
        with contextlib.suppress(FileNotFoundError):
            os.remove(self._LOCK_FILE)

    def _read_available_networks(self) -> List[str]:
        if not os.path.exists(self._NETWORKS_FILE):
            return []

        try:
            with open(self._NETWORKS_FILE) as f:
                network_ids = json.load(f)
                if not isinstance(network_ids, list):
                    logger_no_user_data.error(
                        f'Invalid network buffer file format (expected list, got {type(network_ids).__name__})'
                    )
                    self._backup_corrupted_file()
                    return []
                return network_ids
        except json.JSONDecodeError as error:
            logger_no_user_data.error(f'Corrupted network buffer file: {error}')
            self._backup_corrupted_file()
            return []
        except Exception as error:
            logger_no_user_data.error(f'Failed to read network buffer file: {error}')
            return []

    def _write_available_networks(self, network_ids: List[str]) -> None:
        temp_file = f'{self._NETWORKS_FILE}.tmp'
        try:
            with open(temp_file, 'w') as f:
                json.dump(network_ids, f, indent=2)
                f.flush()
                os.fsync(f.fileno())

            os.replace(temp_file, self._NETWORKS_FILE)
        except Exception as error:
            logger_no_user_data.error(f'Failed to write network buffer file: {error}')
            with contextlib.suppress(FileNotFoundError):
                os.remove(temp_file)
            raise

    def _backup_corrupted_file(self) -> None:
        try:
            timestamp = int(time.time())
            backup_path = f'{self._NETWORKS_FILE}.corrupt-{timestamp}'
            os.rename(self._NETWORKS_FILE, backup_path)
            logger_no_user_data.error(f'Backed up corrupted file to {backup_path}')
        except Exception as error:
            logger_no_user_data.error(f'Failed to backup corrupted file: {error}')

    def allocate_networks(self, job_id: str, count: int) -> List[Network]:
        try:
            self._acquire_lock()

            available_ids = self._read_available_networks()
            allocated: List[Network] = []

            for _ in range(count):
                network = None

                while available_ids and network is None:
                    net_id = available_ids.pop(0)
                    try:
                        network = self._docker.networks.get(net_id)
                        logger_no_user_data.debug(
                            f'Allocated network {network.id} ({network.name}) from buffer for job {job_id}'
                        )
                    except NotFound:
                        logger_no_user_data.warning(
                            f'Network {net_id} in buffer file no longer exists in Docker, skipping'
                        )
                        network = None

                if network is None:
                    logger_no_user_data.debug(f'Buffer exhausted, creating network on-the-fly for job {job_id}')
                    network = self._create_network()

                allocated.append(network)

            self._write_available_networks(available_ids)
            return allocated

        except RuntimeError as error:
            logger_no_user_data.warning(f'Lock acquisition failed: {error}. Creating networks on-the-fly.')
            allocated = []
            for _ in range(count):
                network = self._create_network()
                allocated.append(network)
            return allocated

        finally:
            self._release_lock()

    def fill_buffer(self) -> int:
        try:
            self._acquire_lock()

            available_ids = self._read_available_networks()
            current_count = len(available_ids)
            needed = self.BUFFER_SIZE - current_count

            if needed <= 0:
                logger_no_user_data.debug(
                    f'Buffer already has {current_count} available networks (target: {self.BUFFER_SIZE})'
                )
                return 0

            logger_no_user_data.debug(
                f'Filling buffer: current={current_count}, target={self.BUFFER_SIZE}, creating={needed}'
            )

            created_count = 0
            for _ in range(needed):
                try:
                    network = self._create_network()
                    if network.id:
                        available_ids.append(network.id)
                        created_count += 1
                        logger_no_user_data.debug(f'Created buffer network {network.id} ({created_count}/{needed})')
                    else:
                        logger_no_user_data.error('Created network has no ID, skipping')
                except Exception as error:
                    logger_no_user_data.error(f'Failed to create buffer network: {error}')
                    continue

            self._write_available_networks(available_ids)
            logger_no_user_data.debug(f'Buffer fill complete: created {created_count} networks')
            return created_count

        finally:
            self._release_lock()

    def _create_network(self) -> Network:
        network = _allocate_network_with_retries(
            name_prefix=self.NETWORK_NAME_PREFIX,
            docker_client=self._docker,
            internal=True,
            driver='bridge',
            labels={'biolib-role': 'remote-host-network'},
        )
        return network



import sqlite3
from pathlib import Path

from biolib._shared.types import SqliteV1DatabaseSchema
from biolib.api import client as api_client
from biolib.biolib_api_client import AppGetResponse
from biolib.biolib_api_client.biolib_app_api import _get_app_uri_from_str
from biolib.biolib_api_client.lfs_types import DataRecordVersionInfo


def get_actual_schema(db_path):
    if not db_path.exists():
        raise Exception(f'File {db_path} not found.')
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
    tables = cursor.fetchall()

    actual_schema: SqliteV1DatabaseSchema = {'tables': {}}
    for table in tables:
        table_name = table[0]
        cursor.execute(f'PRAGMA table_info({table_name});')
        columns = cursor.fetchall()
        actual_schema['tables'][table_name] = {'columns': {}}
        for column in columns:
            actual_schema['tables'][table_name]['columns'][column[1]] = {
                'type': column[2],
                'nullable': not bool(column[3]),
            }

        cursor.execute(f'PRAGMA foreign_key_list({table_name});')
        foreign_keys = cursor.fetchall()
        for fk in foreign_keys:
            actual_schema['tables'][table_name]['columns'][fk[3]]['foreign_key'] = {'table': fk[2], 'column': fk[4]}

    conn.close()
    return actual_schema


def verify_schema(specification: SqliteV1DatabaseSchema, actual_schema: SqliteV1DatabaseSchema):
    for table_name, table_spec in specification['tables'].items():
        if table_name not in actual_schema['tables']:
            raise Exception(f"Error: Table '{table_name}' is missing.")

        for column_name, column_spec in table_spec['columns'].items():
            if column_name not in actual_schema['tables'][table_name]['columns']:
                raise Exception(f"Error: Column '{column_name}' in table '{table_name}' is missing.")

            actual_column = actual_schema['tables'][table_name]['columns'][column_name]
            if actual_column['type'] != column_spec['type']:
                raise Exception(
                    f"Error: Column '{column_name}' in table '{table_name}' "
                    "has type '{actual_column['type']}' but expected '{column_spec['type']}'."
                )

            if not actual_column['nullable'] and column_spec.get('nullable', True):
                raise Exception(
                    f"Error: Column '{column_name}' in table '{table_name}' is "
                    'not nullable but should be nullable according to the specification.'
                )

        for column_name, column_spec in table_spec['columns'].items():
            if column_spec.get('foreign_key'):
                foreign_key_spec = column_spec['foreign_key']
                if actual_schema['tables'][table_name]['columns'][column_name].get('foreign_key'):
                    fk = actual_schema['tables'][table_name]['columns'][column_name]['foreign_key']
                    if (
                        fk
                        and foreign_key_spec
                        and fk['table'] == foreign_key_spec['table']
                        and fk['column'] == foreign_key_spec['column']
                    ):
                        raise Exception(
                            f"Error: Column '{column_name}' in table '{table_name}' does "
                            'not have the correct foreign key constraint.'
                        )
                else:
                    raise Exception(
                        f"Error: Column '{column_name}' in table '{table_name}' does "
                        'not have a foreign key constraint.'
                    )


def get_data_record_state_from_uri(uri) -> 'DataRecordVersionInfo':
    normalized_uri = _get_app_uri_from_str(uri)
    app_response: AppGetResponse = api_client.get(path='/app/', params={'uri': normalized_uri}).json()
    resource_uri = app_response['app_version']['app_uri']
    if app_response['app']['type'] != 'data-record':
        raise Exception(f'Resource "{resource_uri}" is not a Data Record')
    return DataRecordVersionInfo(
        resource_uri=app_response['app_version']['app_uri'],
        resource_uuid=app_response['app']['public_id'],
        resource_version_uuid=app_response['app_version']['public_id'],
    )


def validate_sqlite_v1(schema: SqliteV1DatabaseSchema, sqlite_file: Path):
    actual_schema = get_actual_schema(sqlite_file)
    print(schema)
    print(actual_schema)
    verify_schema(specification=schema, actual_schema=actual_schema)



from collections import OrderedDict

from rich.console import Console
from rich.table import Column, Table

from biolib.typing_utils import Any, List


class BioLibTable:
    def __init__(self, columns_to_row_map: OrderedDict, rows: List[Any], title):
        self.title = title
        self.rows = rows
        self.columns_to_row_map = columns_to_row_map
        self.table = self._create_table()

    def _create_table(self) -> Table:
        columns = [Column(header=header, **meta['params']) for header, meta in self.columns_to_row_map.items()]
        table = Table(*columns, title=self.title)
        for row in self.rows:
            row_values: List[str] = []
            for column in self.columns_to_row_map.values():
                keys = column['key'].split('.')
                value = row[keys[0]]
                for key in keys[1:]:
                    if not value or key not in value:
                        continue
                    value = value[key]
                row_values.append(str(value))
            table.add_row(*row_values)
        return table

    def print_table(self):
        console = Console()
        console.print(self.table)



from biolib import api
from biolib.biolib_api_client.api_client import UserTokens
from biolib.typing_utils import TypedDict, Literal


class AuthChallengeCreate(TypedDict):
    token: str


class _AuthChallengeStatus(TypedDict):
    state: Literal['awaiting', 'completed']


class AuthChallengeStatus(_AuthChallengeStatus, total=False):
    user_tokens: UserTokens


class BiolibAuthChallengeApi:

    @staticmethod
    def create_auth_challenge() -> AuthChallengeCreate:
        response = api.client.post(path='/user/auth_challenges/')
        response_dict: AuthChallengeCreate = response.json()
        return response_dict

    @staticmethod
    def get_auth_challenge_status(token: str) -> AuthChallengeStatus:
        response = api.client.get(
            path='/user/auth_challenges/',
            headers={'Auth-Challenge-Token': token},
        )

        response_dict: AuthChallengeStatus = response.json()
        return response_dict



from .result import ResultDict
from .typing import Optional, TypedDict


class ResultCounts(TypedDict):
    cancelled: int
    completed: int
    failed: int
    in_progress: int
    queued: int
    total: int


class DeprecatedExperimentDict(TypedDict):
    # Note: fields on this TypedDict are deprecated
    job_count: int
    job_running_count: int


class ExperimentDict(DeprecatedExperimentDict):
    uuid: Optional[str]
    name: Optional[str]
    account_uuid: Optional[str]
    created_at: Optional[str]
    finished_at: Optional[str]
    last_created_at: Optional[str]
    last_created_result: Optional[ResultDict]
    result_counts: ResultCounts


class ExperimentDetailedDict(ExperimentDict):
    pass



from .experiment import DeprecatedExperimentDict
from .resource_version import ResourceVersionDetailedDict
from .typing import Literal, NotRequired, Optional, TypedDict


class SemanticVersionDict(TypedDict):
    major: int
    minor: int
    patch: int


class ResourceUriDict(TypedDict):
    account_handle_normalized: str
    account_handle: str
    resource_name_normalized: Optional[str]
    resource_name: Optional[str]
    resource_prefix: Optional[str]
    version: Optional[SemanticVersionDict]
    tag: Optional[str]


ResourceTypeLiteral = Literal['app', 'data-record', 'experiment', 'index']


class ResourceDict(TypedDict):
    uuid: str
    uri: str
    name: str
    created_at: str
    description: str
    account_uuid: str


class ResourceDetailedDict(ResourceDict):
    type: ResourceTypeLiteral
    version: NotRequired[ResourceVersionDetailedDict]
    experiment: Optional[DeprecatedExperimentDict]



import os.path
import tarfile
import tempfile

from docker.models.containers import Container  # type: ignore

from biolib.typing_utils import Iterable, Optional


class FileInContainer:
    def __init__(self, container: Container, path_in_container: str, overlay_upper_dir_path: Optional[str]):
        self._container: Container = container
        self._path_on_disk: Optional[str] = overlay_upper_dir_path + path_in_container if overlay_upper_dir_path \
            else None

        self._path_in_container: str = path_in_container
        self._path: str = path_in_container
        self._buffered_file_data: Optional[bytes] = None

    def __repr__(self) -> str:
        return f'FileInContainer({self.path})'

    @property
    def path(self) -> str:
        return self._path

    @path.setter
    def path(self, value: str) -> None:
        self._path = value

    def is_file(self) -> bool:
        if self._path_on_disk:
            return os.path.isfile(self._path_on_disk)
        else:
            tmp_file = self._get_temp_file_from_container_via_tar()
            if tmp_file:
                os.remove(tmp_file)
                return True
            else:
                return False

    def get_data_size_in_bytes(self) -> int:
        if self._path_on_disk:
            return os.path.getsize(self._path_on_disk)
        else:
            tmp_file = self._get_temp_file_from_container_via_tar()
            if tmp_file:
                file_length = os.path.getsize(tmp_file)
                os.remove(tmp_file)
                return file_length
            else:
                return 0

    def get_data_stream(self) -> Iterable[bytes]:
        if self._path_on_disk:
            with open(self._path_on_disk, mode='rb') as file:
                while True:
                    chunk = file.read(1_000_000)
                    if not chunk:
                        return

                    yield chunk
        else:
            tmp_file = self._get_temp_file_from_container_via_tar()
            if not tmp_file:
                yield bytes()
                return
            else:
                file = open(tmp_file, mode='rb')
                while True:
                    chunk = file.read(1_000_000)
                    if not chunk:
                        file.close()
                        os.remove(tmp_file)
                        return

                    yield chunk

    def _get_temp_file_from_container_via_tar(self) -> Optional[str]:
        with tempfile.NamedTemporaryFile(mode='wb', delete=True) as tmp_io:
            stream, _ = self._container.get_archive(path=self._path_in_container)
            for chunk in stream:
                tmp_io.write(chunk)

            tmp_io.seek(0)
            with tarfile.open(tmp_io.name) as tar:
                members = tar.getmembers()
                file_members = [member for member in members if member.isfile()]
                if not file_members:
                    # Path was not a file
                    return None

                assert len(file_members) == 1
                file_obj = tar.extractfile(member=file_members[0])
                if not file_obj:
                    # Path was not a file
                    return None
                tmp_io_2 = tempfile.NamedTemporaryFile(mode='wb', delete=False)
                while True:
                    chunk = file_obj.read(1_000_000)
                    if not chunk:
                        break
                    tmp_io_2.write(chunk)
                tmp_io_2.close()
                return tmp_io_2.name



from .account import AccountDict, AccountDetailedDict
from .account_member import AccountMemberDict
from .app import AppDetailedDict, AppSlimDict
from .data_record import (
    DataRecordDetailedDict,
    DataRecordSlimDict,
    DataRecordTypeDict,
    DataRecordValidationRuleDict,
    SqliteV1Column,
    SqliteV1DatabaseSchema,
    SqliteV1ForeignKey,
    SqliteV1Table,
)
from .experiment import (
    DeprecatedExperimentDict,
    ExperimentDetailedDict,
    ExperimentDict,
    ResultCounts,
)
from .file_node import FileNodeDict, FileZipMetadataDict, ZipFileNodeDict
from .push import PushResponseDict
from .resource import ResourceDetailedDict, ResourceDict, ResourceTypeLiteral, ResourceUriDict, SemanticVersionDict
from .resource_deploy_key import ResourceDeployKeyDict, ResourceDeployKeyWithSecretDict
from .resource_permission import ResourcePermissionDetailedDict, ResourcePermissionDict
from .resource_version import (
    ResourceVersionAssetsDict,
    ResourceVersionDetailedDict,
    ResourceVersionDict,
)
from .result import ResultDetailedDict, ResultDict
from .typing import Optional
from .user import EnterpriseSettingsDict, UserDetailedDict, UserDict

__all__ = [
    'AccountDetailedDict',
    'AccountDict',
    'AccountMemberDict',
    'AppDetailedDict',
    'AppSlimDict',
    'DataRecordDetailedDict',
    'DataRecordSlimDict',
    'DataRecordTypeDict',
    'DataRecordValidationRuleDict',
    'DeprecatedExperimentDict',
    'EnterpriseSettingsDict',
    'ExperimentDetailedDict',
    'ExperimentDict',
    'FileNodeDict',
    'FileZipMetadataDict',
    'Optional',
    'PushResponseDict',
    'ResourceDeployKeyDict',
    'ResourceDeployKeyWithSecretDict',
    'ResourceDetailedDict',
    'ResourceDict',
    'ResourceTypeLiteral',
    'ResourcePermissionDetailedDict',
    'ResourcePermissionDict',
    'ResourceUriDict',
    'ResourceVersionAssetsDict',
    'ResourceVersionDetailedDict',
    'ResourceVersionDict',
    'ResultCounts',
    'ResultDetailedDict',
    'ResultDict',
    'SemanticVersionDict',
    'SqliteV1Column',
    'SqliteV1DatabaseSchema',
    'SqliteV1ForeignKey',
    'SqliteV1Table',
    'UserDetailedDict',
    'UserDict',
    'ZipFileNodeDict',
]



import io
import ipaddress
import tarfile
import time
from urllib.parse import urlparse

from docker.models.containers import Container
from docker.models.networks import Network
from docker.types import EndpointConfig

from biolib import utils
from biolib._internal.utils import base64_encode_string
from biolib.biolib_api_client import BiolibApiClient, RemoteHost
from biolib.biolib_api_client.job_types import CreatedJobDict
from biolib.biolib_docker_client import BiolibDockerClient
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger_no_user_data
from biolib.compute_node.cloud_utils import CloudUtils
from biolib.compute_node.utils import BIOLIB_PROXY_NETWORK_NAME
from biolib.compute_node.webserver.proxy_utils import get_biolib_nginx_proxy_image
from biolib.typing_utils import Dict, List, Optional


def get_static_ip_from_network(network: Network, offset: int = 2) -> str:
    ipam_config = network.attrs['IPAM']['Config']
    if not ipam_config:
        raise BioLibError(f'Network {network.name} has no IPAM configuration')

    subnet_str = ipam_config[0]['Subnet']
    subnet = ipaddress.ip_network(subnet_str, strict=False)

    static_ip = str(subnet.network_address + offset)

    return static_ip


# Prepare for remote hosts with specified port
class RemoteHostExtended(RemoteHost):
    ports: List[int]


class RemoteHostMapping:
    def __init__(self, hostname: str, ports: List[int], network: Network, static_ip: str):
        self.hostname = hostname
        self.ports = ports
        self.network = network
        self.static_ip = static_ip


class RemoteHostProxy:
    def __init__(
        self,
        remote_host_mappings: List[RemoteHostMapping],
        job: CreatedJobDict,
        app_caller_network: Optional[Network] = None,
    ):
        self._remote_host_mappings = remote_host_mappings
        self._app_caller_network = app_caller_network
        self.is_app_caller_proxy = app_caller_network is not None

        if not job:
            raise Exception('RemoteHostProxy missing argument "job"')

        self._job = job
        suffix = '-AppCallerProxy' if app_caller_network else ''
        self._name = f'biolib-remote-host-proxy-{self._job_uuid}{suffix}'
        self._container: Optional[Container] = None
        self._docker = BiolibDockerClient().get_docker_client()

    @property
    def _job_uuid(self) -> str:
        return self._job['uuid']

    def get_hostname_to_ip_mapping(self) -> Dict[str, str]:
        return {mapping.hostname: mapping.static_ip for mapping in self._remote_host_mappings}

    def get_remote_host_networks(self) -> List[Network]:
        networks = [mapping.network for mapping in self._remote_host_mappings]
        return networks

    def get_ip_address_on_network(self, network: Network) -> str:
        if not self._container:
            raise BioLibError('RemoteHostProxy not yet started')

        container_networks = self._container.attrs['NetworkSettings']['Networks']
        if network.name in container_networks:
            ip_address: str = container_networks[network.name]['IPAddress']
            if not ip_address:
                raise BioLibError(f'No IP address found for network {network.name}')
            return ip_address

        raise BioLibError(f'RemoteHostProxy not connected to network {network.name}')

    def start(self) -> None:
        docker = BiolibDockerClient.get_docker_client()

        networking_config: Optional[Dict[str, EndpointConfig]] = (
            None
            if not self.is_app_caller_proxy
            else {
                BIOLIB_PROXY_NETWORK_NAME: docker.api.create_endpoint_config(
                    aliases=[f'biolib-app-caller-proxy-{self._job_uuid}']
                )
            }
        )

        for index in range(3):
            logger_no_user_data.debug(f'Attempt {index} at creating RemoteHostProxy container "{self._name}"...')
            try:
                self._container = docker.containers.create(
                    detach=True,
                    image=get_biolib_nginx_proxy_image(),
                    name=self._name,
                    network=BIOLIB_PROXY_NETWORK_NAME,
                    networking_config=networking_config,
                )
                break
            except Exception as error:
                logger_no_user_data.exception(f'Failed to create container "{self._name}" hit error: {error}')

            logger_no_user_data.debug('Sleeping before re-trying container creation...')
            time.sleep(3)

        if not self._container or not self._container.id:
            raise BioLibError(f'Exceeded re-try limit for creating container {self._name}')

        for mapping in self._remote_host_mappings:
            mapping.network.connect(self._container.id, ipv4_address=mapping.static_ip)
            logger_no_user_data.debug(
                f'Connected proxy to network {mapping.network.name} with static IP {mapping.static_ip}'
            )

        if self._app_caller_network:
            self._app_caller_network.connect(self._container.id)
            logger_no_user_data.debug(f'Connected app caller proxy to network {self._app_caller_network.name}')

        self._write_nginx_config_to_container()

        self._container.start()

        proxy_is_ready = False
        for retry_count in range(1, 5):
            time.sleep(0.5 * retry_count)
            # Use the container logs as a health check.
            # By using logs instead of a http endpoint on the NGINX we avoid publishing a port of container to the host
            container_logs = self._container.logs()
            if b'ready for start up\n' in container_logs or b'start worker process ' in container_logs:
                proxy_is_ready = True
                break

        if not proxy_is_ready:
            self.terminate()
            raise Exception('RemoteHostProxy did not start properly')

        self._container.reload()

    def terminate(self):
        # TODO: Implement nice error handling in this method

        if self._container:
            self._container.remove(force=True)

    def _write_nginx_config_to_container(self) -> None:
        if not self._container:
            raise Exception('RemoteHostProxy container not defined when attempting to write NGINX config')

        docker = BiolibDockerClient.get_docker_client()
        upstream_hostname = urlparse(BiolibApiClient.get().base_url).hostname
        if self.is_app_caller_proxy:
            if not utils.IS_RUNNING_IN_CLOUD:
                raise BioLibError('Calling apps inside apps is not supported in local compute environment')

            logger_no_user_data.debug(f'Job "{self._job_uuid}" writing config for and starting App Caller Proxy...')
            config = CloudUtils.get_webserver_config()
            compute_node_uuid = config['compute_node_info']['public_id']
            compute_node_auth_token = config['compute_node_info']['auth_token']

            # TODO: Get access_token from new API class instead
            access_token = BiolibApiClient.get().access_token
            bearer_token = f'Bearer {access_token}' if access_token else ''

            user_uuid = self._job['user_id']
            if user_uuid:
                biolib_index_basic_auth = f'biolib_user|{user_uuid.replace("-", "_")}:cloud-{compute_node_auth_token}'
                biolib_index_auth_header_value = f'Basic {base64_encode_string(biolib_index_basic_auth)}'
            else:
                biolib_index_auth_header_value = ''

            nginx_config = f"""
events {{
  worker_connections  1024;
}}

http {{
    client_max_body_size 0;
    map $request_method $bearer_token_on_post {{
        POST       "{bearer_token}";
        default    "";
    }}

    map $request_method $bearer_token_on_get {{
        GET        "{bearer_token}";
        default    "";
    }}

    map $request_method $bearer_token_on_patch_and_get {{
        PATCH      "{bearer_token}";
        GET        "{bearer_token}";
        default    "";
    }}

    server {{
        listen       80;
        resolver 127.0.0.11 ipv6=off valid=30s;
        set $upstream_hostname {upstream_hostname};

        location ~* "^/api/jobs/cloud/(?<job_id>[a-z0-9-]{{36}})/status/$" {{
            proxy_pass               https://$upstream_hostname/api/jobs/cloud/$job_id/status/;
            proxy_set_header         authorization $bearer_token_on_get;
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~* "^/api/jobs/cloud/$" {{
            # Note: Using $1 here as URI part from regex must be used for proxy_pass
            proxy_pass               https://$upstream_hostname/api/jobs/cloud/$1;
            proxy_set_header         authorization $bearer_token_on_post;
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~* "^/api/jobs/(?<job_id>[a-z0-9-]{{36}})/storage/input/start_upload/$" {{
            proxy_pass               https://$upstream_hostname/api/jobs/$job_id/storage/input/start_upload/;
            proxy_set_header         authorization "";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~* "^/api/jobs/(?<job_id>[a-z0-9-]{{36}})/storage/input/presigned_upload_url/$" {{
            proxy_pass               https://$upstream_hostname/api/jobs/$job_id/storage/input/presigned_upload_url/$is_args$args;
            proxy_set_header         authorization "";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~* "^/api/jobs/(?<job_id>[a-z0-9-]{{36}})/storage/input/complete_upload/$" {{
            proxy_pass               https://$upstream_hostname/api/jobs/$job_id/storage/input/complete_upload/;
            proxy_set_header         authorization "";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~* "^/api/jobs/(?<job_id>[a-z0-9-]{{36}})/main_result/$" {{
            proxy_pass                  https://$upstream_hostname/api/jobs/$job_id/main_result/;
            proxy_set_header            authorization "";
            proxy_set_header            cookie "";
            proxy_pass_request_headers  on;
            proxy_ssl_server_name       on;
        }}

        location ~* "^/api/jobs/(?<job_id>[a-z0-9-]{{36}})/$" {{
            proxy_pass               https://$upstream_hostname/api/jobs/$job_id/;
            proxy_set_header         authorization $bearer_token_on_patch_and_get;
            proxy_set_header         caller-job-uuid "{self._job_uuid}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~* "^/api/jobs/create_job_with_data/$" {{
            # Note: Using $1 here as URI part from regex must be used for proxy_pass
            proxy_pass               https://$upstream_hostname/api/jobs/create_job_with_data/$1;
            proxy_set_header         authorization $bearer_token_on_post;
            proxy_set_header         caller-job-uuid "{self._job_uuid}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~* "^/api/jobs/$" {{
            # Note: Using $1 here as URI part from regex must be used for proxy_pass
            proxy_pass               https://$upstream_hostname/api/jobs/$1;
            proxy_set_header         authorization $bearer_token_on_post;
            proxy_set_header         caller-job-uuid "{self._job_uuid}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~ "^/api/jobs/{self._job_uuid}/notes/$" {{
            # Note: Using $1 here as URI part from regex must be used for proxy_pass
            proxy_pass               https://$upstream_hostname/api/jobs/{self._job_uuid}/notes/$1;
            proxy_set_header         authorization "";
            proxy_set_header         job-auth-token "";
            proxy_set_header         compute-node-auth-token "{compute_node_auth_token}";
            proxy_set_header         compute-node-uuid "{compute_node_uuid}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~ "^/api/auth/oauth-token-exchange/$" {{
            # Note: Using $1 here as URI part from regex must be used for proxy_pass
            proxy_pass               https://$upstream_hostname/api/auth/oauth-token-exchange/$1;
            proxy_set_header         authorization "";
            proxy_set_header         compute-node-auth-token "{compute_node_auth_token}";
            proxy_set_header         job-uuid "{self._job_uuid}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /api/lfs/ {{
            proxy_pass               https://$upstream_hostname$request_uri;
            proxy_set_header         authorization "";
            proxy_set_header         compute-node-auth-token "{compute_node_auth_token}";
            proxy_set_header         job-uuid "{self._job_uuid}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /api/app/ {{
            proxy_pass               https://$upstream_hostname$request_uri;
            proxy_set_header         authorization "";
            proxy_set_header         compute-node-auth-token "{compute_node_auth_token}";
            proxy_set_header         job-uuid "{self._job_uuid}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /api/resources/data-records/ {{
            proxy_pass               https://$upstream_hostname$request_uri;
            proxy_set_header         authorization "";
            proxy_set_header         compute-node-auth-token "{compute_node_auth_token}";
            proxy_set_header         job-uuid "{self._job_uuid}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /api/resources/versions/ {{
            proxy_pass               https://$upstream_hostname$request_uri;
            proxy_set_header         authorization "";
            proxy_set_header         compute-node-auth-token "{compute_node_auth_token}";
            proxy_set_header         job-uuid "{self._job_uuid}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /api/proxy/files/data-record-versions/ {{
            proxy_pass               https://$upstream_hostname$request_uri;
            proxy_set_header         authorization "";
            proxy_set_header         compute-node-auth-token "{compute_node_auth_token}";
            proxy_set_header         job-uuid "{self._job_uuid}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /api/proxy/index/ {{
            proxy_pass               https://$upstream_hostname$request_uri;
            proxy_set_header         authorization "{biolib_index_auth_header_value}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~* "^/api/accounts/(?<account_id>[a-z0-9-]{{36}})/metrics/jobs/$" {{
            proxy_pass               https://$upstream_hostname/api/accounts/$account_id/metrics/jobs/$is_args$args;
            proxy_set_header         authorization "";
            proxy_set_header         compute-node-auth-token "{compute_node_auth_token}";
            proxy_set_header         job-uuid "{self._job_uuid}";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /api/ {{
            proxy_pass               https://$upstream_hostname$request_uri;
            proxy_set_header         authorization "";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /proxy/storage/job-storage/ {{
            proxy_pass               https://$upstream_hostname$request_uri;
            proxy_set_header         authorization "";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /proxy/storage/lfs/versions/ {{
            proxy_pass               https://$upstream_hostname$request_uri;
            proxy_set_header         authorization "";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /proxy/cloud/ {{
            proxy_pass               https://$upstream_hostname$request_uri;
            proxy_set_header         authorization "";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location / {{
            return 404 "Not found";
        }}
    }}

    server {{
        listen       1080;
        resolver 127.0.0.11 ipv6=off valid=30s;

        if ($http_biolib_result_uuid != "{self._job_uuid}") {{
            return 403 "Invalid or missing biolib-result-uuid header";
        }}

        if ($http_biolib_result_port = "") {{
            return 400 "Missing biolib-result-port header";
        }}

        location / {{
            proxy_pass               http://main:$http_biolib_result_port$request_uri;
            proxy_set_header         biolib-result-uuid "";
            proxy_set_header         biolib-result-port "";
            proxy_pass_request_headers on;
        }}
    }}
}}
"""
        else:
            port_to_mappings: Dict[int, List[RemoteHostMapping]] = {}
            for mapping in self._remote_host_mappings:
                for port in mapping.ports:
                    if port not in port_to_mappings:
                        port_to_mappings[port] = []
                    port_to_mappings[port].append(mapping)

            nginx_config = """
events {}
error_log /dev/stdout info;
stream {
    resolver 127.0.0.11 valid=30s;"""

            for port, mappings in port_to_mappings.items():
                nginx_config += f"""
    map $server_addr $backend_{port} {{"""
                for mapping in mappings:
                    nginx_config += f'\n        {mapping.static_ip} {mapping.hostname}:{port};'

                nginx_config += f"""
    }}
    server {{
        listen 0.0.0.0:{port};
        proxy_pass $backend_{port};
    }}
    server {{
        listen 0.0.0.0:{port} udp;
        proxy_pass $backend_{port};
    }}"""

            nginx_config += """
}
"""

        nginx_config_bytes = nginx_config.encode()
        tarfile_in_memory = io.BytesIO()
        with tarfile.open(fileobj=tarfile_in_memory, mode='w:gz') as tar:
            info = tarfile.TarInfo('/nginx.conf')
            info.size = len(nginx_config_bytes)
            tar.addfile(info, io.BytesIO(nginx_config_bytes))

        tarfile_bytes = tarfile_in_memory.getvalue()
        tarfile_in_memory.close()
        docker.api.put_archive(self._container.id, '/etc/nginx', tarfile_bytes)



from .typing import TypedDict


class ResourceDeployKeyDict(TypedDict):
    uuid: str
    created_at: str
    name: str


class ResourceDeployKeyWithSecretDict(ResourceDeployKeyDict):
    secret_key: str



import hashlib
import ipaddress
import uuid
from typing import Dict, Optional, cast

from docker.errors import APIError
from docker.models.networks import Network
from docker.types import IPAMConfig, IPAMPool

from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger_no_user_data
from biolib.compute_node.remote_host_proxy import get_static_ip_from_network


def _iter_network_subnets(existing_network):
    ipam_config = existing_network.attrs.get('IPAM', {}).get('Config', [])
    for cfg in ipam_config:
        subnet_str = cfg.get('Subnet')
        if not subnet_str:
            continue
        try:
            yield ipaddress.ip_network(subnet_str, strict=False)
        except ValueError:
            continue


def _find_overlap(candidate_network, existing_networks):
    for existing in existing_networks:
        for subnet in _iter_network_subnets(existing):
            if candidate_network.overlaps(subnet):
                return existing, str(subnet)
    return None


def _allocate_network_with_retries(
    name_prefix: str,
    docker_client,
    internal: bool = True,
    driver: str = 'bridge',
    max_attempts: int = 10,
    labels: Optional[Dict[str, str]] = None,
) -> Network:
    base_network = ipaddress.ip_network('172.28.0.0/16', strict=False)

    suffix = uuid.uuid4().hex
    full_name = f'{name_prefix}{suffix}'
    name_hash = int(hashlib.sha256(full_name.encode()).hexdigest(), 16)
    starting_offset = name_hash % 256

    for attempt in range(max_attempts):
        offset = (starting_offset + attempt) % 256

        if base_network.prefixlen == 16:
            third_octet = offset
            candidate_subnet = f'{base_network.network_address.exploded.rsplit(".", 2)[0]}.{third_octet}.0/24'
        else:
            candidate_subnet = f'{base_network.network_address.exploded.rsplit(".", 1)[0]}.{offset}.0/24'

        candidate_network = ipaddress.ip_network(candidate_subnet, strict=False)

        existing_networks = docker_client.networks.list()
        overlap = _find_overlap(candidate_network, existing_networks)
        if overlap:
            existing_network, existing_subnet = overlap
            logger_no_user_data.debug(
                f'Subnet {candidate_subnet} conflicts with existing network '
                f'{existing_network.name} ({existing_subnet}), trying next candidate'
            )
            continue

        ipam_pool = IPAMPool(subnet=candidate_subnet)
        computed_ipam_config = IPAMConfig(pool_configs=[ipam_pool])

        try:
            network = cast(
                Network,
                docker_client.networks.create(
                    name=full_name,
                    internal=internal,
                    driver=driver,
                    ipam=computed_ipam_config,
                    labels=labels or {},
                ),
            )
            static_ip = get_static_ip_from_network(network, offset=2)
            logger_no_user_data.debug(
                f'Created network {full_name} with subnet {candidate_subnet} and static IP {static_ip}'
            )
            return network
        except APIError as api_error:
            logger_no_user_data.debug(
                f'Network creation failed with Docker API error for subnet {candidate_subnet}: {api_error}, '
                f'trying next candidate (attempt {attempt + 1}/{max_attempts})'
            )
            continue

    raise BioLibError(
        f'Failed to allocate and create network {full_name} after {max_attempts} attempts. ' f'Base CIDR: 172.28.0.0/16'
    )
